# -*- coding: utf-8 -*-

"""User interface for command line script

License: See LICENSE file.

"""

import argparse

from . import __version__
from .core import process_backup


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i", "--input-dir", help="Input directory", required=True
    )
    parser.add_argument(
        "-o", "--output-dir", help="Output directory", required=True
    )
    parser.add_argument(
        "-V",
        "--version",
        help="Show version and exit",
        action="version",
        version=__version__,
    )
    return parser.parse_args()


def main():
    args = parse_args()
    process_backup(args.input_dir, args.output_dir)



# -*- coding: utf-8 -*-

"""Core functionality

License: See LICENSE file.

"""

import base64
import binascii
import datetime as dt
import logging
import os
import shutil
import sqlite3
import uuid

from typing import List

from .__version__ import __version__
from .addressbook import Addressbook
from .addressbook import make_addressbook
from .dbproto import StructuredGroupCall
from .dbproto import StructuredGroupDataV1
from .dbproto import StructuredGroupDataV2
from .dbproto import StructuredMemberRole
from .dbproto import StructuredMentions
from .dbproto import StructuredReactions
from .exceptions import DatabaseNotFound
from .exceptions import DatabaseVersionNotFound
from .html import dump_thread
from .models import Attachment
from .models import GroupCallData
from .models import GroupUpdateData
from .models import MemberInfo
from .models import Mention
from .models import MMSMessageRecord
from .models import Quote
from .models import Reaction
from .models import Recipient
from .models import SMSMessageRecord
from .models import Thread
from .types import is_group_call
from .types import is_group_ctrl
from .types import is_group_v2_data
from .versioninfo import VersionInfo

logger = logging.getLogger(__name__)


def check_backup(backup_dir) -> VersionInfo:
    """Check that we have the necessary files and return VersionInfo"""
    if not os.path.join(backup_dir, "database.sqlite"):
        raise DatabaseNotFound
    if not os.path.join(backup_dir, "DatabaseVersion.sbf"):
        raise DatabaseVersionNotFound
    with open(os.path.join(backup_dir, "DatabaseVersion.sbf"), "r") as fp:
        version_str = fp.read()

    version = version_str.split(":")[-1].strip()
    versioninfo = VersionInfo(version)

    if not versioninfo.is_tested_version():
        logger.warn(
            f"This database version is untested, please report errors."
        )
    return versioninfo


def get_color(db, recipient_id):
    """Extract recipient color from the database"""
    query = db.execute(
        "SELECT color FROM recipient_preferences WHERE recipient_ids=?",
        (recipient_id,),
    )
    color = query.fetchone()[0]
    return color


def get_sms_records(db, thread, addressbook):
    """Collect all the SMS records for a given thread"""
    sms_records = []
    sms_qry = db.execute(
        "SELECT _id, address, date, date_sent, body, type, "
        "delivery_receipt_count, read_receipt_count "
        "FROM sms WHERE thread_id=?",
        (thread._id,),
    )
    qry_res = sms_qry.fetchall()
    for (
        _id,
        address,
        date,
        date_sent,
        body,
        _type,
        delivery_receipt_count,
        read_receipt_count,
    ) in qry_res:

        data = get_data_from_body(_type, body, addressbook, _id)
        sms_auth = addressbook.get_recipient_by_address(str(address))
        sms = SMSMessageRecord(
            _id=_id,
            data=data,
            delivery_receipt_count=delivery_receipt_count,
            read_receipt_count=read_receipt_count,
            addressRecipient=sms_auth,
            recipient=thread.recipient,
            dateSent=date_sent,
            dateReceived=date,
            threadId=thread._id,
            body=body,
            _type=_type,
        )
        sms_records.append(sms)
    return sms_records


def get_attachment_filename(_id, unique_id, backup_dir, thread_dir):
    """Get the absolute path of an attachment, warn if it doesn't exist"""
    fname = f"Attachment_{_id}_{unique_id}.bin"
    source = os.path.abspath(os.path.join(backup_dir, fname))
    if not os.path.exists(source):
        logger.warn(
            f"Couldn't find attachment '{source}'. "
            "Maybe it was deleted or never downloaded?"
        )
        return None

    # Copying here is a bit of a side-effect
    target_dir = os.path.abspath(os.path.join(thread_dir, "attachments"))
    os.makedirs(target_dir, exist_ok=True)
    target = os.path.join(target_dir, fname)
    shutil.copy(source, target)
    url = "/".join([".", "attachments", fname])
    return url


def add_mms_attachments(db, mms, backup_dir, thread_dir):
    """Add all attachment objects to MMS message"""
    qry = db.execute(
        "SELECT _id, ct, unique_id, voice_note, width, height, quote "
        "FROM part WHERE mid=?",
        (mms._id,),
    )
    for _id, ct, unique_id, voice_note, width, height, quote in qry:
        a = Attachment(
            contentType=ct,
            unique_id=unique_id,
            fileName=get_attachment_filename(
                _id, unique_id, backup_dir, thread_dir
            ),
            voiceNote=voice_note,
            width=width,
            height=height,
            quote=quote,
        )
        mms.attachments.append(a)


def decode_body(body):
    """Decode a base64-encoded message body."""
    try:
        return base64.b64decode(body)
    except (TypeError, ValueError, binascii.Error) as e:
        logger.warn(f"Failed to decode body for message '{body}': {str(e)}")
        return None


def get_group_call_data(rawbody, addressbook, mid):
    """Get the data for a group call."""

    if not rawbody:
        return None

    try:
        structured_call = StructuredGroupCall.loads(rawbody)
    except (ValueError, IndexError, TypeError) as e:
        logger.warn(
            f"Failed to load group call data for message {mid}:\n"
            f"Error message: {str(e)}"
        )
        return None

    timestamp = dt.datetime.fromtimestamp(structured_call.when // 1000)
    timestamp = timestamp.replace(
        microsecond=(structured_call.when % 1000) * 1000
    )
    recipient = addressbook.get_recipient_by_uuid(structured_call.by)
    if recipient:
        initiator = recipient.name

    group_call_data = GroupCallData(
        initiator=initiator,
        timestamp=timestamp,
    )

    return group_call_data


def get_group_update_data_v1(rawbody, addressbook, mid):
    """Get the data for a Group V1 update.

    There are two lists of members:
    - One list by structures with telephone numbers and/or UUIDs
    - Another list by telephone numbers

    Some groups use the first (sometimes without filling the UUID field),
    some use the second, and some both.

    Merge both lists and indicate for each member if they were found by phone
    or by UUID (preferable)."""

    if not rawbody:
        return None

    try:
        structured_group_data = StructuredGroupDataV1.loads(rawbody)
    except (ValueError, IndexError, TypeError) as e:
        logger.warn(
            f"Failed to load group update data (v1) for message {mid}:\n"
            f"Error message: {str(e)}"
        )
        return None

    members = dict()
    for member in structured_group_data.members:
        name = None
        match_from_phone = False
        if member.uuid:
            name = addressbook.get_recipient_by_uuid(member.uuid).name
        elif member.phone:
            name = addressbook.get_recipient_by_phone(member.phone).name
            match_from_phone = True

        member_info = MemberInfo(
            name=name,
            phone=member.phone,
            match_from_phone=match_from_phone,
            admin=False,
        )
        members[member.phone] = member_info

    for phone_member in structured_group_data.phone_members:
        if not members.get(phone_member):
            name = addressbook.get_recipient_by_phone(phone_member).name
            member_info = MemberInfo(
                name=name,
                phone=phone_member,
                match_from_phone=True,
                admin=False,
            )
            members[phone_member] = member_info

    members_list = list()
    for key, value in members.items():
        members_list.append(value)

    group_update_data = GroupUpdateData(
        group_name=structured_group_data.group_name,
        change_by=None,
        members=members_list,
    )
    return group_update_data


def get_member_by_raw_uuid(raw_uuid: bytes, what: str, addressbook, mid: str):
    """Find a recipient from a binary UUID. Output their name from the
    addressbook or the textual UUID if not found."""

    try:
        text_uuid = str(uuid.UUID(bytes=raw_uuid))
    except ValueError as e:
        logger.warn(f"Failed to parse {what} UUID for message {mid}: {str(e)}")
        return None

    recipient = addressbook.get_recipient_by_uuid(text_uuid)
    member_name = recipient.name if recipient else text_uuid
    return member_name


def get_group_update_data_v2(rawbody, addressbook, mid):
    """Get the data for a Group V2 update.

    Group V2 updates use UUIDs exclusively to identify members. The update
    messages contain the old state, the new state, and a description of the
    changes. Parse the change description to print out:
    - The person who made the change (editor)
    - A new group name
    - Added (new) members
    - Deleted members

    Also parse the new state to print out the resulting list of members."""

    if not rawbody:
        return None

    try:
        structured_group_data = StructuredGroupDataV2.loads(rawbody)
    except (ValueError, IndexError, TypeError) as e:
        logger.warn(
            f"Failed to load group update data (v2) for message {mid}:\n"
            f"Error message: {str(e)}"
        )
        return None

    change = structured_group_data.change
    deleted_members = []
    new_members = []
    members = []
    change_by = None
    editor = None
    if (not change is None) and (not change.by is None):
        editor = get_member_by_raw_uuid(
            change.by, "update editor", addressbook, mid
        )
        if editor:
            change_by = MemberInfo(
                name=editor,
                phone=None,
                match_from_phone=False,
                admin=False,
            )

    if (not change is None) and (not change.new_members is None):
        for member in change.new_members:
            name = (
                get_member_by_raw_uuid(
                    member.uuid, "new member", addressbook, mid
                )
                or "Unknown"
            )
            admin = member.role == StructuredMemberRole.MEMBER_ROLE_ADMIN
            new_members.append(
                MemberInfo(
                    name=name, phone=None, match_from_phone=False, admin=admin
                )
            )

    if (not change is None) and (not change.deleted_members is None):
        for member in change.deleted_members:
            name = (
                get_member_by_raw_uuid(
                    member, "deleted member", addressbook, mid
                )
                or "Unknown"
            )
            deleted_members.append(
                MemberInfo(
                    name=name, phone=None, match_from_phone=False, admin=False
                )
            )

    state = structured_group_data.state
    for member in state.members:
        name = (
            get_member_by_raw_uuid(member.uuid, "member", addressbook, mid)
            or "Unknown"
        )
        admin = member.role == StructuredMemberRole.MEMBER_ROLE_ADMIN
        members.append(
            MemberInfo(
                name=name, phone=None, match_from_phone=False, admin=admin
            )
        )

    if (change is None) or (change.new_title is None):
        group_name = None
    else:
        group_name = change.new_title.value

    group_update_data = GroupUpdateData(
        group_name=group_name,
        change_by=change_by,
        members=members,
        new_members=new_members,
        deleted_members=deleted_members,
    )
    return group_update_data


def get_data_from_body(_type, body, addressbook, mid):
    """Decode data in the message body and provide a structured representation."""
    data = None
    if is_group_call(_type):
        data = get_group_call_data(decode_body(body), addressbook, mid)
    elif is_group_ctrl(_type):
        if is_group_v2_data(_type):
            data = get_group_update_data_v2(
                decode_body(body), addressbook, mid
            )
        else:
            data = get_group_update_data_v1(
                decode_body(body), addressbook, mid
            )

    return data


def get_mms_mentions(encoded_mentions, addressbook, mid):
    """Decode mentions encoded in a SQL blob."""
    mentions = {}
    if not encoded_mentions:
        return mentions

    try:
        structured_mentions = StructuredMentions.loads(encoded_mentions)
    except (ValueError, IndexError, TypeError) as e:
        logger.warn(
            f"Failed to load quote mentions for message {mid}:\n"
            f"Error message: {str(e)}"
        )
        return mentions

    for structured_mention in structured_mentions.mentions:
        recipient = addressbook.get_recipient_by_uuid(
            structured_mention.who_uuid
        )
        name = recipient.name
        mention = Mention(
            mention_id=-1, name=name, length=structured_mention.length
        )
        range_start = (
            0 if structured_mention.start is None else structured_mention.start
        )
        mentions[range_start] = mention

    return mentions


def get_mms_reactions(encoded_reactions, addressbook, mid):
    """Decode reactions encoded in a SQL blob."""
    reactions = []
    if not encoded_reactions:
        return reactions

    try:
        structured_reactions = StructuredReactions.loads(encoded_reactions)
    except (ValueError, IndexError, TypeError) as e:
        logger.warn(
            f"Failed to load reactions for message {mid}:\n"
            f"Error message: {str(e)}"
        )
        return reactions

    for structured_reaction in structured_reactions.reactions:
        recipient = addressbook.get_recipient_by_address(
            str(structured_reaction.who)
        )
        reaction = Reaction(
            recipient=recipient,
            what=structured_reaction.what,
            time_sent=dt.datetime.fromtimestamp(
                structured_reaction.time_sent // 1000
            ),
            time_received=dt.datetime.fromtimestamp(
                structured_reaction.time_received // 1000
            ),
        )
        reaction.time_sent = reaction.time_sent.replace(
            microsecond=(structured_reaction.time_sent % 1000) * 1000
        )
        reaction.time_received = reaction.time_received.replace(
            microsecond=(structured_reaction.time_received % 1000) * 1000
        )
        reactions.append(reaction)

    return reactions


def get_mms_records(
    db, thread, addressbook, backup_dir, thread_dir, versioninfo
):
    """Collect all MMS records for a given thread"""
    mms_records = []

    reaction_expr = versioninfo.get_reactions_query_column()
    quote_mentions_expr = versioninfo.get_quote_mentions_query_column()
    viewed_receipt_count_expr = versioninfo.get_viewed_receipt_count_column()

    qry = db.execute(
        "SELECT _id, address, date, date_received, body, quote_id, "
        f"quote_author, quote_body, {quote_mentions_expr}, msg_box, {reaction_expr}, "
        f"delivery_receipt_count, read_receipt_count, {viewed_receipt_count_expr} "
        "FROM mms WHERE thread_id=?",
        (thread._id,),
    )
    qry_res = qry.fetchall()
    for (
        _id,
        address,
        date,
        date_received,
        body,
        quote_id,
        quote_author,
        quote_body,
        quote_mentions,
        msg_box,
        reactions,
        delivery_receipt_count,
        read_receipt_count,
        viewed_receipt_count,
    ) in qry_res:
        quote = get_mms_quote(
            addressbook,
            quote_id,
            quote_author,
            quote_body,
            quote_mentions,
            _id,
        )

        decoded_reactions = get_mms_reactions(reactions, addressbook, _id)

        data = get_data_from_body(msg_box, body, addressbook, _id)
        mms_auth = addressbook.get_recipient_by_address(str(address))
        mms = MMSMessageRecord(
            _id=_id,
            data=data,
            delivery_receipt_count=delivery_receipt_count,
            read_receipt_count=read_receipt_count,
            addressRecipient=mms_auth,
            recipient=thread.recipient,
            dateSent=date,
            dateReceived=date_received,
            threadId=thread._id,
            body=body,
            quote=quote,
            attachments=[],
            reactions=decoded_reactions,
            _type=msg_box,
            viewed_receipt_count=viewed_receipt_count,
        )
        mms_records.append(mms)

    for mms in mms_records:
        add_mms_attachments(db, mms, backup_dir, thread_dir)

    return mms_records


def get_mms_quote(
    addressbook, quote_id, quote_author, quote_body, quote_mentions, mid
):
    """Retrieve quote (replied message) from a MMS message."""
    quote = None
    if quote_id:
        quote_auth = addressbook.get_recipient_by_address(quote_author)
        decoded_mentions = get_mms_mentions(quote_mentions, addressbook, mid)
        quote = Quote(
            _id=quote_id,
            author=quote_auth,
            text=quote_body,
            mentions=decoded_mentions,
        )
    return quote


def get_mentions(db, addressbook, thread_id, versioninfo):
    """Retrieve all mentions in the DB for the requested thread into a dictionary."""
    mentions = {}

    if versioninfo.are_mentions_supported():
        query = db.execute(
            "SELECT _id, message_id, recipient_id, range_start, range_length "
            "FROM mention WHERE thread_id=?",
            (thread_id,),
        )
        mentions_data = query.fetchall()

        for (
            _id,
            message_id,
            recipient_id,
            range_start,
            range_length,
        ) in mentions_data:
            name = addressbook.get_recipient_by_address(str(recipient_id)).name
            mention = Mention(
                mention_id=_id,
                name=name,
                length=range_length,
            )
            if not message_id in mentions.keys():
                mentions[message_id] = {}
            mentions[message_id][range_start] = mention

    return mentions


def get_members(
    db: sqlite3.Cursor,
    addressbook: Addressbook,
    thread_id: int,
    versioninfo: VersionInfo,
) -> List[Recipient]:
    """Retrieve the thread members from the database

    Returns
    -------
    members: List[Recipient]
        A list of Recipients for each member in the group.
    """
    thread_rid_column = versioninfo.get_thread_recipient_id_column()
    if versioninfo.is_addressbook_using_rids():
        query = db.execute(
            "SELECT r._id, g.members "
            "FROM thread t "
            "LEFT JOIN recipient r "
            f"ON t.{thread_rid_column} = r._id "
            "LEFT JOIN groups g "
            "ON g.group_id = r.group_id "
            "WHERE t._id = :thread_id",
            {"thread_id": thread_id},
        )
        query_result = query.fetchall()
        recipient_id, thread_members = query_result[0]
    else:
        query = db.execute(
            "SELECT t.recipient_ids, g.members "
            "FROM thread t "
            "LEFT JOIN groups g "
            "ON t.recipient_ids = g.group_id "
            "WHERE t._id = :thread_id",
            {"thread_id": thread_id},
        )
        query_result = query.fetchall()
        recipient_id, thread_members = query_result[0]

    if not thread_members is None:
        member_addresses = thread_members.split(",")
        members = []
        for address in member_addresses:
            recipient = addressbook.get_recipient_by_address(address)
            members.append(recipient)
    else:
        members = [addressbook.get_recipient_by_address(recipient_id)]
    return members


def populate_thread(
    db, thread, addressbook, backup_dir, thread_dir, versioninfo=None
):
    """Populate a thread with all corresponding messages"""
    sms_records = get_sms_records(db, thread, addressbook)
    mms_records = get_mms_records(
        db,
        thread,
        addressbook,
        backup_dir,
        thread_dir,
        versioninfo,
    )
    thread.sms = sms_records
    thread.mms = mms_records
    thread.mentions = get_mentions(db, addressbook, thread._id, versioninfo)
    thread.members = get_members(db, addressbook, thread._id, versioninfo)


def process_backup(backup_dir, output_dir):
    """Main functionality to convert database into HTML"""

    logger.info(f"This is signal2html version {__version__}")

    # Verify backup and open database
    versioninfo = check_backup(backup_dir)
    db_file = os.path.join(backup_dir, "database.sqlite")
    db_conn = sqlite3.connect(db_file)
    db = db_conn.cursor()

    # Get and index all contact and group names
    addressbook = make_addressbook(db, versioninfo)

    # Start by getting the Threads from the database
    recipient_id_expr = versioninfo.get_thread_recipient_id_column()

    query = db.execute(f"SELECT _id, {recipient_id_expr} FROM thread")
    threads = query.fetchall()

    # Combine the recipient objects and the thread info into Thread objects
    for (_id, recipient_id) in threads:
        recipient = addressbook.get_recipient_by_address(str(recipient_id))
        if recipient is None:
            logger.warn(f"No recipient with address {recipient_id}")

        t = Thread(_id=_id, recipient=recipient)
        thread_dir = t.get_thread_dir(output_dir, make_dir=False)
        populate_thread(
            db, t, addressbook, backup_dir, thread_dir, versioninfo=versioninfo
        )
        dump_thread(t, output_dir)

    db.close()



# -*- coding: utf-8 -*-

"""Command line interface

License: See LICENSE file.
"""


def main():
    import logging
    import sys

    from .ui import main

    logging.basicConfig(
        datefmt="%Y-%m-%d %H:%M:%S",
        format="%(asctime)s | %(levelname)s - %(message)s",
        level=logging.INFO,
    )
    sys.exit(main())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Module for linkifying urls in messages

License: See LICENSE file

"""

import logging

from linkify_it import LinkifyIt
from linkify_it.tlds import TLDS

logger = logging.getLogger(__name__)


def linkify(message: str) -> str:
    """Replace text URLs in message with HTML links"""
    linkifier = LinkifyIt().tlds(TLDS)

    # Pre-test first for efficiency
    if not linkifier.pretest(message):
        return message

    # Test if a link is present
    if not linkifier.test(message):
        return message

    # Find links in message
    matches = linkifier.match(message)
    if not matches:
        return message

    logger.debug(f"Replacing urls in message:\n{message}")

    # Construct new message
    new_message = ""
    idx = 0
    for match in matches:
        new_message += message[idx : match.index]
        new_message += f'<a href="{match.url}" target="_blank">{match.raw}</a>'
        idx = match.last_index
    new_message += message[idx:]

    logger.debug(f"Replaced urls in message:\n{new_message}")
    return new_message



# -*- coding: utf-8 -*-

"""Colors extracted from Signal source

License: See LICENSE file.

"""

import logging
import random

logger = logging.getLogger(__name__)

CRIMSON = "#CC163D"
CRIMSON_TINT = "#EDA6AE"
CRIMSON_SHADE = "#8A0F29"
VERMILLION = "#C73800"
VERMILLION_TINT = "#EBA78E"
VERMILLION_SHADE = "#872600"
BURLAP = "#746C53"
BURLAP_TINT = "#C4B997"
BURLAP_SHADE = "#58513C"
FOREST = "#3B7845"
FOREST_TINT = "#8FCC9A"
FOREST_SHADE = "#2B5934"
WINTERGREEN = "#1C8260"
WINTERGREEN_TINT = "#9BCFBD"
WINTERGREEN_SHADE = "#36544A"
TEAL = "#067589"
TEAL_TINT = "#A5CAD5"
TEAL_SHADE = "#055968"
BLUE = "#336BA3"
BLUE_TINT = "#ADC8E1"
BLUE_SHADE = "#285480"
INDIGO = "#5951C8"
INDIGO_TINT = "#C2C1E7"
INDIGO_SHADE = "#4840A0"
VIOLET = "#862CAF"
VIOLET_TINT = "#CDADDC"
VIOLET_SHADE = "#6B248A"
PLUMB = "#A23474"
PLUMB_TINT = "#DCB2CA"
PLUMB_SHADE = "#881B5B"
TAUPE = "#895D66"
TAUPE_TINT = "#CFB5BB"
TAUPE_SHADE = "#6A4E54"
STEEL = "#6B6B78"
STEEL_TINT = "#BEBEC6"
STEEL_SHADE = "#5A5A63"
ULTRAMARINE = "#2C6BED"
ULTRAMARINE_TINT = "#B0C8F9"
ULTRAMARINE_SHADE = "#1851B4"
GROUP = "#2C6BED"
GROUP_TINT = "#B0C8F9"
GROUP_SHADE = "#1851B4"

COLORMAP = {
    "red": CRIMSON,
    "deep_orange": CRIMSON,
    "orange": VERMILLION,
    "amber": VERMILLION,
    "brown": BURLAP,
    "yellow": BURLAP,
    "pink": PLUMB,
    "purple": VIOLET,
    "deep_purple": VIOLET,
    "indigo": INDIGO,
    "blue": BLUE,
    "light_blue": BLUE,
    "cyan": TEAL,
    "teal": TEAL,
    "green": FOREST,
    "light_green": WINTERGREEN,
    "lime": WINTERGREEN,
    "blue_grey": TAUPE,
    "grey": STEEL,
    "ultramarine": ULTRAMARINE,
    "group_color": GROUP,
}

# Extracted from:
# https://github.com/signalapp/Signal-Android/blob/master/app/src/main/java/org/thoughtcrime/securesms/conversation/colors/AvatarColor.java
# Note that Signal uses ColorInts, which are AARRGGBB. We strip off the alpha.
AVATAR_COLORS = {
    "C000": "#D00B0B",
    "C010": "#C72A0A",
    "C020": "#B34209",
    "C030": "#9C5711",
    "C040": "#866118",
    "C050": "#76681E",
    "C060": "#6C6C13",
    "C070": "#5E6E0C",
    "C080": "#507406",
    "C090": "#3D7406",
    "C100": "#2D7906",
    "C110": "#1A7906",
    "C120": "#067906",
    "C130": "#067919",
    "C140": "#06792D",
    "C150": "#067940",
    "C160": "#067953",
    "C170": "#067462",
    "C180": "#067474",
    "C190": "#077288",
    "C200": "#086DA0",
    "C210": "#0A69C7",
    "C220": "#0D59F2",
    "C230": "#3454F4",
    "C240": "#5151F6",
    "C250": "#6447F5",
    "C260": "#7A3DF5",
    "C270": "#8F2AF4",
    "C280": "#A20CED",
    "C290": "#AF0BD0",
    "C300": "#B80AB8",
    "C310": "#C20AA3",
    "C320": "#C70A88",
    "C330": "#CB0B6B",
    "C340": "#D00B4D",
    "C350": "#D00B2C",
    "crimson": "#CF163E",
    "vermilion": "#C73F0A",
    "burlap ": "#6F6A58",
    "forest ": "#3B7845",
    "wintergreen": "#1D8663",
    "teal": "#077D92",
    "blue": "#336BA3",
    "indigo ": "#6058CA",
    "violet ": "#9932CB",
    "plum": "#AA377A",
    "taupe": "#8F616A",
    "steel": "#71717F",
    "unknown": "#71717F",
}


def list_colors():
    return sorted(set(list(COLORMAP.keys()) + list(AVATAR_COLORS.keys())))


def get_color(name):
    color = COLORMAP.get(name, None) or AVATAR_COLORS.get(name, None)
    if not color is None:
        return color
    logger.warn(f"Unknown color: {name}, using fallback color instead.")
    return AVATAR_COLORS["unknown"]


def get_random_color():
    return random.choice(list(COLORMAP.keys()))



# -*- coding: utf-8 -*-

"""
Models for storing Signal backup objects.

These are heavily inspired by the database models of the Signal Android app, 
but only the necessary fields are kept.

License: See LICENSE file.

"""

import os

from abc import ABCMeta
from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from re import sub
from unicodedata import normalize

from typing import Dict
from typing import List


@dataclass
class Recipient:
    rid: int
    name: str
    color: str
    isgroup: bool
    phone: str
    uuid: str

    def __hash__(self):
        return hash(self.rid)


@dataclass
class Mention:
    mention_id: int
    name: str
    length: int


@dataclass
class Quote:
    _id: int
    author: Recipient
    text: str
    mentions: Dict[int, Mention] = field(default_factory=lambda: {})


@dataclass
class Attachment:
    contentType: str
    fileName: str
    voiceNote: bool
    width: int
    height: int
    quote: bool
    unique_id: str


@dataclass
class GroupCallData:
    initiator: str
    timestamp: datetime


@dataclass
class MemberInfo:
    name: str
    phone: str
    match_from_phone: bool
    admin: bool


@dataclass
class GroupUpdateData:
    group_name: str
    change_by: MemberInfo
    members: List[MemberInfo] = field(default_factory=list)
    new_members: List[MemberInfo] = field(default_factory=list)
    deleted_members: List[MemberInfo] = field(default_factory=list)


@dataclass
class DisplayRecord(metaclass=ABCMeta):
    addressRecipient: Recipient  # Recipient corresponding to address field
    recipient: Recipient
    dateSent: int
    dateReceived: int
    threadId: int
    body: str
    _type: int


@dataclass
class MessageRecord(DisplayRecord):
    _id: int
    data: any
    delivery_receipt_count: int
    read_receipt_count: int


@dataclass
class Reaction:
    recipient: Recipient
    what: str
    time_sent: datetime
    time_received: datetime


@dataclass
class MMSMessageRecord(MessageRecord):
    quote: Quote
    attachments: List[Attachment]
    reactions: List[Reaction]
    viewed_receipt_count: int


@dataclass
class SMSMessageRecord(MessageRecord):
    pass


@dataclass
class Thread:
    _id: int
    recipient: Recipient
    mms: List[MMSMessageRecord] = field(default_factory=lambda: [])
    sms: List[SMSMessageRecord] = field(default_factory=lambda: [])
    mentions: Dict[int, Dict[int, Mention]] = field(default_factory=lambda: {})
    members: List[Recipient] = field(default_factory=lambda: [])

    @property
    def is_group(self) -> bool:
        return self.recipient.isgroup

    @property
    def sanephone(self):
        """Return a sanitized phone number suitable for use as filename, and fallback on rid.

        NOTE: Phone numbers can be alphanumerical characters especially coming over SMS
        Different strings can still clash in principle if they sanitize to the same string.
        """
        if self.recipient.phone:
            return self._sanitize(self.recipient.phone)
        return "#" + str(self.recipient.rid)

    @property
    def name(self):
        """Return the raw name or other useful identifier, suitable for display."""
        return self.recipient.name.strip()

    @property
    def sanename(self):
        """Return a sanitized name or other useful identifier, suitable for use
        as filename, and fallback on rid."""
        if self.recipient.name:
            return self._sanitize(self.recipient.name)
        return "#" + str(self.recipient.rid)

    def _sanitize(self, text):
        """Sanitize text to use as filename"""
        clean = normalize("NFKC", text.strip())
        clean = clean.lstrip(".#")
        clean = sub("[^]\\w!@#$%^&'`.=+{}~()[-]", "_", clean)
        clean = clean.rstrip("_")
        return clean

    def get_thread_dir(self, output_dir: str, make_dir=True) -> str:
        return os.path.dirname(self.get_path(output_dir, make_dir=make_dir))

    def get_path(self, output_dir: str, make_dir=True) -> str:
        """Return a path for a thread and try to be clever about merging
        contacts. Optionally create the contact directory."""
        dirname = self.sanename
        # Use phone number to distinguish threads from the same contact,
        # except for groups, which do not have a phone number.
        filename = f"{self.sanename if self.is_group else self.sanephone}.html"
        path = os.path.join(output_dir, dirname, filename)
        i = 2
        while os.path.exists(path):
            if self.is_group:
                dirname = f"{self.sanename}_{i}"
            else:
                filename = f"{self.sanephone}_{i}.html"
            path = os.path.join(output_dir, dirname, filename)
            i += 1
        if make_dir:
            os.makedirs(os.path.dirname(path), exist_ok=True)
        return path



# -*- coding: utf-8 -*-

"""Utilities for dealing with message types

These are extracted from the Signal Android app source code.

License: See LICENSE file.

"""

from enum import Enum

BASE_TYPE_MASK = 0x1F

INCOMING_AUDIO_CALL_TYPE = 1
OUTGOING_AUDIO_CALL_TYPE = 2
MISSED_AUDIO_CALL_TYPE = 3
JOINED_TYPE = 4
UNSUPPORTED_MESSAGE_TYPE = 5
INVALID_MESSAGE_TYPE = 6
MISSED_VIDEO_CALL_TYPE = 8
INCOMING_VIDEO_CALL_TYPE = 10
OUTGOING_VIDEO_CALL_TYPE = 11
GROUP_CALL_TYPE = 12

KEY_UPDATE_TYPE_BIT = 0x200

GROUP_CTRL_TYPE_BIT = 0x10000
GROUP_V2_DATA_TYPE_BIT = 0x80000

SECURE_BIT = 0x800000

BASE_INBOX_TYPE = 20
BASE_OUTBOX_TYPE = 21
BASE_SENDING_TYPE = 22
BASE_SENT_TYPE = 23
BASE_SENT_FAILED_TYPE = 24
BASE_PENDING_SECURE_SMS_FALLBACK = 25
BASE_PENDING_INSECURE_SMS_FALLBACK = 26
BASE_DRAFT_TYPE = 27

OUTGOING_MESSAGE_TYPES = [
    BASE_OUTBOX_TYPE,
    BASE_SENT_TYPE,
    BASE_SENDING_TYPE,
    BASE_SENT_FAILED_TYPE,
    BASE_PENDING_SECURE_SMS_FALLBACK,
    BASE_PENDING_INSECURE_SMS_FALLBACK,
]


class DisplayType(Enum):
    DISPLAY_TYPE_NONE = 0
    DISPLAY_TYPE_PENDING = 1
    DISPLAY_TYPE_SENT = 2
    DISPLAY_TYPE_FAILED = 3
    DISPLAY_TYPE_DELIVERED = 4
    DISPLAY_TYPE_READ = 5

    @classmethod
    def from_state(
        cls, _type: int, is_delivered: bool = False, is_read: bool = False
    ):
        if not is_outgoing_message_type(_type):
            return cls.DISPLAY_TYPE_NONE

        if is_read:
            return cls.DISPLAY_TYPE_READ

        if is_delivered:
            return cls.DISPLAY_TYPE_DELIVERED

        base_type = _type & BASE_TYPE_MASK

        if base_type == BASE_SENT_FAILED_TYPE:
            return cls.DISPLAY_TYPE_FAILED

        if base_type == BASE_SENT_TYPE:
            return cls.DISPLAY_TYPE_SENT

        if (
            base_type == BASE_OUTBOX_TYPE
            or base_type == BASE_SENDING_TYPE
            or base_type == BASE_PENDING_SECURE_SMS_FALLBACK
            or base_type == BASE_PENDING_INSECURE_SMS_FALLBACK
        ):
            return cls.DISPLAY_TYPE_PENDING

        return cls.DISPLAY_TYPE_NONE


def is_outgoing_message_type(_type):
    for outgoingType in OUTGOING_MESSAGE_TYPES:
        if _type & BASE_TYPE_MASK == outgoingType:
            return True
    return False


def is_inbox_type(_type):
    return _type & BASE_TYPE_MASK == BASE_INBOX_TYPE


def is_incoming_call(_type):
    return _type in (INCOMING_AUDIO_CALL_TYPE, INCOMING_VIDEO_CALL_TYPE)


def is_outgoing_call(_type):
    return _type in (OUTGOING_AUDIO_CALL_TYPE, OUTGOING_VIDEO_CALL_TYPE)


def is_missed_call(_type):
    return _type in (MISSED_AUDIO_CALL_TYPE, MISSED_VIDEO_CALL_TYPE)


def is_video_call(_type):
    return _type in (
        INCOMING_VIDEO_CALL_TYPE,
        OUTGOING_VIDEO_CALL_TYPE,
        MISSED_VIDEO_CALL_TYPE,
    )


def is_group_call(_type):
    return _type == GROUP_CALL_TYPE


def is_key_update(_type):
    return _type & KEY_UPDATE_TYPE_BIT == KEY_UPDATE_TYPE_BIT


def is_secure(_type):
    return _type & SECURE_BIT == SECURE_BIT


def is_group_ctrl(_type):
    return _type & GROUP_CTRL_TYPE_BIT == GROUP_CTRL_TYPE_BIT


def is_group_v2_data(_type):
    return _type & GROUP_V2_DATA_TYPE_BIT == GROUP_V2_DATA_TYPE_BIT


def is_joined_type(_type):
    return _type & BASE_TYPE_MASK == JOINED_TYPE


def get_named_message_type(_type):
    if is_group_ctrl(_type):
        if is_group_v2_data(_type):
            return "group-update-v2"
        else:
            return "group-update-v1"
    elif is_key_update(_type):
        return "key-update"
    elif is_outgoing_message_type(_type):
        return "outgoing"
    elif is_inbox_type(_type):
        return "incoming"
    elif is_incoming_call(_type):
        return (
            "video-call-incoming" if is_video_call(_type) else "call-incoming"
        )
    elif is_outgoing_call(_type):
        return (
            "video-call-outgoing" if is_video_call(_type) else "call-outgoing"
        )
    elif is_missed_call(_type):
        return "video-call-missed" if is_video_call(_type) else "call-missed"
    elif is_group_call(_type):
        return "group-call"
    elif is_joined_type(_type):
        return "joined"
    return "unknown"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Extraction classes for Signal protocol objects.

These classes are used to extract values in the database encoded as protobuf messages.

Signal protobuf definitions are defined in these directories:

    https://github.com/signalapp/Signal-Android/tree/master/libsignal/service/src/main/proto
    https://github.com/signalapp/Signal-Android/tree/master/app/src/main/proto

License: See LICENSE file.
"""

from dataclasses import dataclass
from enum import IntEnum

from typing import List
from typing import Optional

from pure_protobuf.dataclasses_ import field
from pure_protobuf.dataclasses_ import message
from pure_protobuf.dataclasses_ import optional_field
from pure_protobuf.types import uint32
from pure_protobuf.types import uint64


@message
@dataclass
class StructuredReaction:
    what: str = optional_field(1)
    who: Optional[uint64] = optional_field(2)
    time_sent: Optional[uint64] = optional_field(3)
    time_received: Optional[uint64] = optional_field(4)


@message
@dataclass
class StructuredReactions:
    reactions: List[StructuredReaction] = field(1, default_factory=list)


@message
@dataclass
class StructuredMention:
    start: uint32 = optional_field(1)
    length: uint32 = optional_field(2)
    who_uuid: str = optional_field(3)


@message
@dataclass
class StructuredMentions:
    mentions: List[StructuredMention] = field(1, default_factory=list)


@message
@dataclass
class StructuredGroupCall:
    by: str = optional_field(2)
    when: uint64 = optional_field(3)


@message
@dataclass
class StructuredGroupMember:
    uuid: str = optional_field(1)
    phone: str = optional_field(2)


@message
@dataclass
class StructuredGroupDataV1:
    group_name: str = optional_field(3)
    phone_members: List[str] = field(4, default_factory=list)
    members: List[StructuredGroupMember] = field(6, default_factory=list)


class StructuredMemberRole(IntEnum):
    MEMBER_ROLE_UNKNOWN = 0
    MEMBER_ROLE_DEFAULT = 1
    MEMBER_ROLE_ADMIN = 2


@message
@dataclass
class StructuredDecryptedMember:
    uuid: bytes = field(1)
    role: StructuredMemberRole = field(2)


@message
@dataclass
class StructuredDecryptedString:
    value: str = field(1)


@message
@dataclass
class StructuredGroupV2Change:
    by: bytes = optional_field(1)
    new_members: List[StructuredDecryptedMember] = field(
        3, default_factory=list
    )
    deleted_members: List[bytes] = field(4, default_factory=list)
    new_title: StructuredDecryptedString = optional_field(10)


@message
@dataclass
class StructuredGroupV2State:
    title: str = optional_field(2)
    rev: uint32 = optional_field(6)
    members: List[StructuredDecryptedMember] = field(7, default_factory=list)


@message
@dataclass
class StructuredGroupDataV2:
    change: StructuredGroupV2Change = optional_field(2)
    state: StructuredGroupV2State = optional_field(3)



# -*- coding: utf-8 -*-

"""Code for writing out the HTML

License: See LICENSE file.

"""

import datetime as dt
import logging

from types import SimpleNamespace as ns

from emoji import emoji_list
from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import select_autoescape

from .html_colors import get_color
from .html_colors import list_colors
from .linkify import linkify
from .models import MMSMessageRecord
from .models import Thread
from .types import DisplayType
from .types import get_named_message_type
from .types import is_group_call
from .types import is_group_ctrl
from .types import is_inbox_type
from .types import is_incoming_call
from .types import is_joined_type
from .types import is_key_update
from .types import is_missed_call
from .types import is_outgoing_call
from .types import is_secure

logger = logging.getLogger(__name__)


def is_all_emoji(body):
    """Check if a message is non-empty and only contains emoji"""
    body = body.replace(" ", "").replace("\ufe0f", "")
    return len(emoji_list(body)) == len(body) and len(body) > 0


def format_message(body, mentions=None):
    """Format message by processing all characters.

    - Wrap emoji in <span> for styling them
    - Escape special HTML chars
    """
    if body is None:
        return None

    if mentions is None:
        mentions = {}

    emoji_pos = emoji_list(body)
    new_body = ""
    emoji_lookup = {p["location"]: p["emoji"] for p in emoji_pos}
    skip = 0
    for i, c in enumerate(body):
        if skip > 0:
            # Skip additional characters from multi-character emoji
            skip = skip - 1
        elif i in emoji_lookup:
            new_body += "<span class='msg-emoji'>%s</span>" % emoji_lookup[i]
            skip = len(emoji_lookup[i]) - 1
        elif c == "&":
            new_body += "&amp;"
        elif c == "<":
            new_body += "&lt;"
        elif c == ">":
            new_body += "&gt;"
        elif c == "\ufffc":  # Object replacement character
            mention = mentions.get(i)
            if mention:
                new_body += (
                    "<span class='msg-mention'>@%s</span>"
                    % format_message(mention.name)
                )
                skip = (
                    mention.length - 1
                )  # Not clear in what case this is not 1
            else:
                new_body += c
        else:
            new_body += c

    new_body = linkify(new_body)
    return new_body


def format_member_list(header: str, member_list):
    """Return a list of printable group members belonging to a category (e.g. new members)."""
    people = ns()
    people.header = header
    people.members = list()
    for member in member_list:
        designation = ""
        if not member.match_from_phone:
            if member.phone:
                designation = f"{format_message(member.name)} ({format_message(member.phone)})"
            else:
                designation = f"{format_message(member.name)}"
        else:
            if member.name is None or member.name == member.phone:
                designation = f"{format_message(member.phone)}"
            else:
                designation = f"{format_message(member.phone)} ~ {format_message(member.name)}"

        if member.admin:
            designation += " (admin)"

        people.members.append(designation)

    return people


def format_event_data_group_update(data):
    """Return a structure describing a group update event."""
    event_data = ns()
    event_data.member_lists = list()

    event_data.header = "Group update"
    if data.change_by:
        event_data.header += " by " + format_message(data.change_by.name)

    if data.group_name:
        event_data.name = data.group_name

    if data.new_members and len(data.new_members) > 0:
        member_list = format_member_list("New members:", data.new_members)
        event_data.member_lists.append(member_list)

    if data.deleted_members and len(data.deleted_members) > 0:
        member_list = format_member_list(
            "Deleted members:", data.deleted_members
        )
        event_data.member_lists.append(member_list)

    if data.members and len(data.members) > 0:
        member_list = format_member_list("Members:", data.members)
        event_data.member_lists.append(member_list)

    return event_data


def dump_thread(thread: Thread, output_dir: str):
    """Write a Thread instance to a HTML page in the output directory"""

    # Combine and sort the messages
    messages = thread.mms + thread.sms
    messages.sort(key=lambda mr: mr.dateSent)

    # Find the template
    env = Environment(
        loader=PackageLoader("signal2html", "templates"),
        autoescape=select_autoescape(["html", "xml"]),
    )
    template = env.get_template("thread.html")

    # Create the message color CSS (depends on individuals)
    group_color_css = ""
    msg_css = ".msg-sender-%i { /* recipient id: %5s */ background: %s;}\n"
    if thread.is_group:
        group_recipients = set(m.addressRecipient for m in messages)
        sender_idx = {r: k for k, r in enumerate(group_recipients)}
        colors_used = []
        group_colors = set(ar.color for ar in sender_idx)
        for ar, idx in sender_idx.items():
            if ar.isgroup:
                continue

            # ensure colors are unique, even if they're not in Signal
            ar_color = ar.color
            if ar_color in colors_used:
                color = next(
                    (c for c in list_colors() if not c in group_colors),
                    None,
                )
                ar_color = ar.color if color is None else color
            group_color_css += msg_css % (
                idx,
                ar.rid,
                get_color(ar_color),
            )
            colors_used.append(ar.color)
    else:
        # Retrieve sender info from an incoming message, if any
        firstInbox = next(
            (m for m in messages if is_inbox_type(m._type)), None
        )
        if firstInbox:
            clr = firstInbox.addressRecipient.color
            clr = "teal" if clr is None else clr
            group_color_css += msg_css % (
                0,
                firstInbox.addressRecipient.rid,
                get_color(clr),
            )

    # Create a simplified dict for each message
    prev_date = None
    simple_messages = []
    for msg in messages:

        if is_joined_type(msg._type):
            continue

        # Add a "date change" message when to mark the date
        date_sent = dt.datetime.fromtimestamp(msg.dateSent // 1000)
        date_sent = date_sent.replace(microsecond=(msg.dateSent % 1000) * 1000)
        if prev_date is None or date_sent.date() != prev_date:
            prev_date = date_sent.date()
            out = {
                "date_msg": True,
                "body": date_sent.strftime("%a, %b %d, %Y"),
            }
            simple_messages.append(out)

        # Handle event messages (calls, group changes)
        is_event = False
        event_data = None
        if is_incoming_call(msg._type):
            is_event = True
            event_data = format_message(thread.name)
        elif is_outgoing_call(msg._type):
            is_event = True
        elif is_missed_call(msg._type):
            is_event = True
        elif is_group_call(msg._type):
            is_event = True
            if msg.data is not None:
                if msg.data.initiator:
                    event_data = format_message(msg.data.initiator)
            else:
                logger.warn(f"Group call for {msg._id} without data")
        elif is_key_update(msg._type):
            is_event = True
            event_data = format_message(msg.addressRecipient.name)
        elif is_group_ctrl(msg._type) and not msg.data is None:
            is_event = True
            event_data = format_event_data_group_update(
                msg.data
            )  # "Group update (v2)"

        # Deal with quoted messages
        quote = {}
        if isinstance(msg, MMSMessageRecord) and msg.quote:
            quote_author_id = msg.quote.author.rid
            quote_author_name = msg.quote.author.name
            if quote_author_id == quote_author_name:
                name = "You"
            else:
                name = quote_author_name
            quote = {
                "name": name,
                "body": format_message(msg.quote.text, msg.quote.mentions),
                "attachments": [],
            }

        # Clean up message body
        body = "" if msg.body is None else msg.body
        if isinstance(msg, MMSMessageRecord):
            all_emoji = not msg.quote and is_all_emoji(body)
        else:
            all_emoji = is_all_emoji(body)

        # Skip HTML/mentions clean-up if this is an event (formatting included in event)
        if not is_event:
            body = format_message(body, thread.mentions.get(msg._id))

        send_state = str(
            DisplayType.from_state(
                msg._type,
                msg.delivery_receipt_count > 0,
                msg.read_receipt_count > 0,
            )
        )
        send_state = send_state[
            send_state.index(".") + 1 :
        ]  # A bit hackish, StrEnum would be better (Python 3.10)

        # Create message dictionary
        aR = msg.addressRecipient
        out = {
            "isAllEmoji": all_emoji,
            "isGroup": thread.is_group,
            "isCall": is_event,
            "type": get_named_message_type(msg._type),
            "body": body,
            "event_data": event_data if is_event else None,
            "date": date_sent,
            "attachments": [],
            "id": msg._id,
            "name": aR.name,
            "secure": is_secure(msg._type) or is_event,
            "send_state": send_state,
            "delivery_receipt_count": msg.delivery_receipt_count,
            "read_receipt_count": msg.read_receipt_count,
            "sender_idx": sender_idx[aR] if thread.is_group else "0",
            "quote": quote,
            "reactions": [],
        }

        # Add attachments and reactions
        if isinstance(msg, MMSMessageRecord):
            for a in msg.attachments:
                if a.quote:
                    out["quote"]["attachments"].append(a)
                else:
                    out["attachments"].append(a)

            for r in msg.reactions:
                out["reactions"].append(
                    {
                        "recipient_id": r.recipient.rid,
                        "name": r.recipient.name,
                        "what": r.what,
                        "time_sent": r.time_sent,
                        "time_received": r.time_received,
                    }
                )

        simple_messages.append(out)

    if not simple_messages:
        return

    if thread.is_group:
        count = len(thread.members)
        subtitle = f"{count} member" if count == 1 else f"{count} members"
    else:
        subtitle = thread.sanephone

    html = template.render(
        thread_name=thread.name,
        thread_subtitle=subtitle,
        messages=simple_messages,
        group_color_css=group_color_css,
        date_time_format="%b %d, %H:%M",
    )
    output_file = thread.get_path(output_dir)
    with open(output_file, "w", encoding="utf-8") as fp:
        fp.write(html)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import io
import os

from setuptools import find_packages
from setuptools import setup

# Package meta-data.
NAME = "signal2html"
DESCRIPTION = "Convert your Signal backup to pretty HTML"
URL = "https://github.com/GjjvdBurg/signal2html"
EMAIL = "gertjanvandenburg@gmail.com"
AUTHOR = "Gertjan van den Burg"
REQUIRES_PYTHON = ">=3.6.0"
VERSION = None

# What packages are required for this module to be executed?
REQUIRED = [
    "emoji>=2.0",
    "jinja2",
    'dataclasses;python_version=="3.6"',
    "pure-protobuf",
    "linkify-it-py",
]

docs_require = []
test_require = []
dev_require = ["green", "black", "isort"]

# What packages are optional?
EXTRAS = {
    "docs": docs_require,
    "tests": test_require,
    "dev": docs_require + test_require + dev_require,
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(
        exclude=["tests", "*.tests", "*.tests.*", "tests.*"]
    ),
    entry_points={
        "console_scripts": ["signal2html=signal2html.__main__:main"],
    },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Class grouping version-specific information from the signal database.

This file in the Signal source can be used to determine when features were 
introduced:

    https://github.com/signalapp/Signal-Android/blob/master/app/src/main/java/org/thoughtcrime/securesms/database/helpers/SignalDatabaseMigrations.kt
    
License: See LICENSE file.

"""

import logging

logger = logging.getLogger(__name__)


class VersionInfo(object):
    def __init__(self, version):
        self.version = int(version)
        logger.info(f"Found Signal database version: {version}.")

    def is_tested_version(self) -> bool:
        """Returns whether the database version has been tested.

        Testing and pull requests welcome."""

        return self.version in [18, 23, 65, 80, 89, 110]

    def is_addressbook_using_rids(self) -> bool:
        """Returns whether the contacts are structured using recipient IDs.

        Previous versions referred to contacts using their phone numbers or a
        special group ID."""

        return self.version >= 24

    def get_reactions_query_column(self) -> str:
        """Returns a SQL expression to retrieve reactions to MMS messages."""

        return "reactions" if self.version >= 37 else "''"

    def are_mentions_supported(self) -> bool:
        """Returns whether the mentions table is present."""
        return self.version >= 68

    def get_quote_mentions_query_column(self) -> str:
        """Returns a SQL expression to retrieve quote mentions in MMS messages."""

        return "quote_mentions" if self.are_mentions_supported() else "''"

    def get_viewed_receipt_count_column(self) -> str:
        """Returns a SQL expression to retrieve the viewed receipt count of attachments of MMS messages."""

        return "viewed_receipt_count" if self.version >= 83 else "'0'"

    def get_thread_recipient_id_column(self) -> str:
        """Returns SQL expression to retrieve recipient id from thread table"""
        return (
            "thread_recipient_id" if self.version >= 108 else "recipient_ids"
        )



# -*- coding: utf-8 -*-

"""Addressbook functionality

License: See LICENSE file.

"""

import abc
import logging

from .html_colors import get_random_color
from .models import Recipient
from .versioninfo import VersionInfo


class Addressbook(metaclass=abc.ABCMeta):
    """Abstract class that store contacts and groups.

    Note: subclasses must implement at a minimum:

    - `_load_recipients()` to load all recipients
    - `get_recipient_by_address()` to return a specific recipient"""

    def __init__(self, db):
        """Initializes the addressbook and load all known recipients."""
        self.logger = logging.getLogger(__name__)
        self.db = db
        self.rid_to_recipient: dict[str, Recipient] = {}
        self.phone_to_rid: dict[str, str] = {}
        self.uuid_to_rid: dict[str, str] = {}
        self.groups: dict[int, str] = {}

        self._load_groups()
        self._load_recipients()  # Must be implemented by subclass
        self.next_rid = 10000

    @abc.abstractmethod
    def _load_recipients():
        """Load all recipients in the recipient_preferences table."""

    def get_group_title(self, group_id: str) -> str:
        """Retrieves the title of a group given the group_id (long
        hexadecimal-based identifier)."""
        return self.groups.get(group_id)

    @abc.abstractmethod
    def get_recipient_by_address(self, address: str) -> Recipient:
        """Returns a Recipient object that matches the address provided.

        The address is the kind of information found in the address field
        of the mms/sms message tables, but also in recipient_ids of the
        thread table.

        If an address is provided that does not exist in the addressbook,
        it is created on the spot."""

    def get_recipient_by_phone(self, phone: str) -> Recipient:
        """Returns a Recipient object that matches the phone number provided."""
        rid = self.phone_to_rid.get(phone)
        return self.rid_to_recipient.get(rid)

    def get_recipient_by_uuid(self, uuid: str) -> Recipient:
        """Returns a Recipient object that matches the UUID provided."""
        rid = self.uuid_to_rid.get(uuid)
        return self.rid_to_recipient.get(rid)

    def _add_recipient(self, recipient_id, uuid, name, color, isgroup, phone):
        """Adds a recipient to the internal data structures."""
        recipient = Recipient(
            recipient_id,
            name=name,
            color=color,
            isgroup=isgroup,
            phone=phone,
            uuid=uuid,
        )

        self.rid_to_recipient[str(recipient_id)] = recipient
        self.logger.debug(
            f"Adding recipient {str(recipient_id)} and phone {phone}"
        )
        if phone:
            self.phone_to_rid[str(phone)] = str(recipient_id)
        if uuid:
            self.uuid_to_rid[uuid] = str(recipient_id)

        return recipient

    def _get_friendly_name_for_group(self, address: str):
        """Creates a readable group name, either the title or a name derived from the group id."""
        name = self.get_group_title(address)
        if not name:
            gid = self._get_group_id(address)
            if gid:
                return f"Group {gid}"
            else:
                return ""

    def _get_group_id(self, group_id: str) -> str:
        """Gets the integer ID of a group from the Signal database."""
        qry = self.db.execute(
            "SELECT group_id, _id FROM groups WHERE group_id LIKE ?",
            (f"{group_id}",),
        )
        qry_res = qry.fetchone()
        if qry_res:
            return str(qry_res[1])

    def _get_new_rid(self) -> str:
        """Creates a new recipient ID for recipients not in the initial
        addressbook."""
        while self.rid_to_recipient.get(str(self.next_rid)):
            self.next_rid += 1

        return str(self.next_rid)

    def _get_unique_group_id(self, group_id: str) -> str:
        """Given a group ID, returns a unique identifier for the group.

        Group IDs are currently comprised of a type, followed by '!', followed
        by an hexadecimal identifier.

        NOTE: This method currently returns the group id itself, but might be
        used to merge groups that share the same hexadecimal identifier."""
        return group_id

    def _load_groups(self):
        """Loads all group names (a.k.a. titles)."""
        qry = self.db.execute("SELECT group_id, title FROM groups")
        qry_res = qry.fetchall()
        for group_id, title in qry_res:
            self.groups[self._get_unique_group_id(group_id)] = title


class AddressbookV1(Addressbook):
    def get_recipient_by_address(self, address: str) -> Recipient:
        """In this database version, all addresses are directly phone numbers
        or group_id's and creating them might happen if no preferences were
        stored for the particular address."""

        isgroup = self._isgroup(address)
        if isgroup:
            phone = self._get_unique_group_id(address)
        else:
            phone = address

        rid = self.phone_to_rid.get(phone)
        recipient = self.rid_to_recipient.get(rid)

        if recipient is None:
            # Create on the spot
            newrid = self._get_new_rid()
            if isgroup:
                friendly_name = self._get_friendly_name_for_group(phone)
                if friendly_name:
                    self.logger.info(
                        f"Group '{phone}' not in addressbook, adding it as '{friendly_name}'."
                    )
                else:
                    self.logger.warn(
                        f"Group '{phone}' not in addressbook, adding it with new ID {newrid}."
                    )
                return self._add_recipient(
                    newrid, "", friendly_name, get_random_color(), True, phone
                )
            else:
                self.logger.info(
                    f"Recipient with phone '{address}' not in addressbook, adding it."
                )
                return self._add_recipient(
                    newrid, "", address, get_random_color(), False, phone
                )
        else:
            return recipient

    def _isgroup(self, address: str) -> bool:
        """Decides whether an address refers to a group."""
        return address.startswith(
            "__textsecure_group__"
        ) or address.startswith("__signal_mms_group__")

    def _load_recipients(self):
        """Load all recipients in the recipient_preferences table.

        In this version of the database, it is normal for recipients in other
        tables not to be found in this table."""
        qry = self.db.execute(
            "SELECT _id, recipient_ids, system_display_name, color, signal_profile_name "
            "FROM recipient_preferences "
        )
        qry_res = qry.fetchall()

        for (
            recipient_id,
            phone,
            system_display_name,
            color,
            profile_name,
        ) in qry_res:
            isgroup = self._isgroup(phone)
            if isgroup:
                phone = self._get_unique_group_id(phone)
                name = self.get_group_title(phone)
                if name is None:
                    name = self._get_friendly_name_for_group(phone)
                    self.logger.warn(
                        f"Group for recipient {recipient_id} will be named '{name}'."
                    )
            else:
                name = system_display_name or profile_name or phone or ""

            if color is None:
                color = get_random_color()

            self._add_recipient(recipient_id, "", name, color, isgroup, phone)


class AddressbookV2(Addressbook):
    def get_recipient_by_address(self, address: str) -> Recipient:
        """In this database version, all addresses are recipient_id's and
        creating them here is not expected to happen."""

        rid = str(address)
        recipient = self.rid_to_recipient.get(rid)

        if recipient is None:
            # Create on the spot, but not expected to happen
            self.logger.warn(
                f"Recipient with rid {address} not in addressbook, adding it."
            )
            return self._add_recipient(
                rid, "", "", get_random_color(), False, ""
            )
        else:
            return recipient

    def _isgroup(self, group_id) -> bool:
        """Decides whether a group_id refers to a group."""
        return group_id is not None

    def _load_recipients(self):
        """Load all recipients in the recipient table.

        In this version of the database, all recipients references should be
        found in this table."""
        qry = self.db.execute(
            "SELECT _id, group_id, uuid, "
            "phone, "
            "system_display_name, "
            "profile_joined_name, "
            "color "
            "FROM recipient "
        )
        qry_res = qry.fetchall()
        for (
            recipient_id,
            group_id,
            uuid,
            phone,
            system_display_name,
            profile_joined_name,
            color,
        ) in qry_res:
            isgroup = self._isgroup(group_id)
            if isgroup:
                name = self.get_group_title(group_id)
                if name is None:
                    name = self._get_friendly_name_for_group(group_id)
                    if name:
                        self.logger.info(
                            f"Group for recipient {recipient_id} is '{group_id}' and will be called by group id using name '{name}'."
                        )
                    else:
                        self.logger.warn(
                            f"Group for recipient {recipient_id} is '{group_id}' which does not exist."
                        )
            else:
                name = (
                    system_display_name or profile_joined_name or phone or ""
                )

            if color is None:
                color = get_random_color()

            self._add_recipient(
                recipient_id, uuid, name, color, isgroup, phone
            )


def make_addressbook(db, versioninfo) -> Addressbook:
    """Factory function for Addressbook.

    The returned implementation depends on the structure of the Signal database."""
    if versioninfo.is_addressbook_using_rids():
        return AddressbookV2(db)
    return AddressbookV1(db)
